from typing import Any
from unittest.mock import MagicMock
from google.cloud import kms
import pytest
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3

from web3_google_hsm import GCPKmsAccount
from web3_google_hsm.types.ethereum_types import Transaction, Signature

# Test Constants
TEST_ADDRESS = "0x0545640A0EcD6FB6ae94766811F30dCDA4746DFC"
TEST_KEY_PATH = "projects/test-project/locations/global/keyRings/test-ring/cryptoKeys/test-key/cryptoKeyVersions/1"
TEST_PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAE0hPxTjwIf407JpkjCdf9kwVPvGdMOZUq\nGaVPbV4qdocIUoJlxmWoOQeL/mR28cLrRqgn+Uj8HAoman2lndsp3w==\n-----END PUBLIC KEY-----\n"

TEST_MESSAGE = "Hello Ethereum!"

# This is a DER-encoded test signature
TEST_DER_SIGNATURE = bytes.fromhex("3045022100cd497d0da6f4a4b9f78c6cad57fc120fd1ac351d84cf549abd437e15c4bd5b280220703f25964ae85b6885525e35fe5cde08b5f4a47512be5a914d715ee415b9b017")

@pytest.fixture
def web3():
    """Initialize Web3 instance."""
    return Web3(Web3.HTTPProvider("http://localhost:8545"))

@pytest.fixture
def test_address() -> ChecksumAddress:
    """Get a test Ethereum address."""
    return to_checksum_address(TEST_ADDRESS)

@pytest.fixture
def test_message() -> str:
    """Get a test message."""
    return TEST_MESSAGE

@pytest.fixture
def test_signature() -> Signature:
    """Create a test signature."""
    return Signature(
        v=27,
        r=bytes.fromhex("cd497d0da6f4a4b9f78c6cad57fc120fd1ac351d84cf549abd437e15c4bd5b28"),
        s=bytes.fromhex("703f25964ae85b6885525e35fe5cde08b5f4a47512be5a914d715ee415b9b017")
    )

@pytest.fixture
def transaction_dict() -> dict[str, Any]:
    """Create a test transaction dictionary."""
    return {
        "from": TEST_ADDRESS,
        "chain_id": 1,
        "nonce": 0,
        "gas_price": 20_000_000_000,
        "gas_limit": 21000,
        "to": TEST_ADDRESS,
        "value": 1_000_000_000_000_000_000,  # 1 ETH
        "data": "0x"
    }

@pytest.fixture
def test_transaction(transaction_dict: dict[str, Any]) -> Transaction:
    """Create a test transaction."""
    return Transaction.from_dict(transaction_dict)

@pytest.fixture
def mock_kms_client() -> MagicMock:
    """Create a mock KMS client."""
    mock_client = MagicMock(spec=kms.KeyManagementServiceClient)

    # Mock the get_public_key response
    mock_public_key_response = MagicMock()
    mock_public_key_response.pem = TEST_PUBLIC_KEY
    mock_client.get_public_key.return_value = mock_public_key_response

    # Mock the asymmetric_sign response with real test values in DER format
    mock_sign_response = MagicMock()
    mock_sign_response.signature = TEST_DER_SIGNATURE
    mock_client.asymmetric_sign.return_value = mock_sign_response

    return mock_client

@pytest.fixture
def gcp_kms_account(mock_kms_client: MagicMock) -> GCPKmsAccount:
    """Create a GCP KMS account with mocked client."""
    account = GCPKmsAccount(key_path=TEST_KEY_PATH)
    account._client = mock_kms_client
    return account

@pytest.fixture
def funded_account(web3: Web3) -> Any:
    """Create a funded test account using hardhat's first account."""
    return web3.eth.account.from_key(
        "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
    )

@pytest.fixture
def fund_test_account(web3: Web3, funded_account: Any, test_address: ChecksumAddress):
    """Fund the test account with some ETH."""
    tx = {
        'from': funded_account.address,
        'to': test_address,
        'value': web3.to_wei(0.1, 'ether'),
        'gas': 21000,
        'gasPrice': web3.eth.gas_price,
        'nonce': web3.eth.get_transaction_count(funded_account.address),
        'chainId': web3.eth.chain_id
    }

    signed_tx = web3.eth.account.sign_transaction(tx, funded_account.key)
    tx_hash = web3.eth.send_raw_transaction(signed_tx.rawTransaction)
    return web3.eth.wait_for_transaction_receipt(tx_hash)

@pytest.fixture
def address() -> ChecksumAddress:
    """Get a test Ethereum address."""
    return to_checksum_address(TEST_ADDRESS)



"""Unit tests for configuration."""
import pytest
from web3_google_hsm.config import (
    BaseConfig,
    ENV_PROJECT_ID,
    ENV_LOCATION_ID,
    ENV_KEY_RING_ID,
    ENV_KEY_ID,
    ENV_WEB3_PROVIDER_URI,
    DEFAULT_WEB3_PROVIDER_URI,
)
from unittest.mock import patch
from pydantic import ValidationError

# Test Constants
TEST_CONFIG = {
    "project_id": "test-project",
    "location_id": "test-region",
    "key_ring_id": "test-keyring",
    "key_id": "test-key",
}

def test_env_constants():
    """Test that environment variable constants are correct."""
    assert ENV_PROJECT_ID == "GOOGLE_CLOUD_PROJECT"
    assert ENV_LOCATION_ID == "GOOGLE_CLOUD_REGION"
    assert ENV_KEY_RING_ID == "KEY_RING"
    assert ENV_KEY_ID == "KEY_NAME"
    assert ENV_WEB3_PROVIDER_URI == "WEB3_PROVIDER_URI"
    assert DEFAULT_WEB3_PROVIDER_URI == "http://localhost:8545"

def test_config_from_env():
    """Test configuration creation from environment variables."""
    env_vars = {
        ENV_PROJECT_ID: "env-project",
        ENV_LOCATION_ID: "env-region",
        ENV_KEY_RING_ID: "env-keyring",
        ENV_KEY_ID: "env-key",
        ENV_WEB3_PROVIDER_URI: "http://custom:8545"
    }

    with patch.dict("os.environ", env_vars, clear=True):
        config = BaseConfig.from_env()
        assert config.project_id == "env-project"
        assert config.location_id == "env-region"
        assert config.key_ring_id == "env-keyring"
        assert config.key_id == "env-key"
        assert config.web3_provider_uri == "http://custom:8545"

def test_config_from_empty_env():
    """Test configuration creation from empty environment."""
    with patch.dict("os.environ", {}, clear=True):
        with pytest.raises(ValidationError) as exc_info:
            BaseConfig.from_env()
        errors = exc_info.value.errors()
        fields = {"project_id", "location_id", "key_ring_id", "key_id"}
        error_fields = {error["loc"][0] for error in errors}
        assert fields.issubset(error_fields)

def test_config_from_partial_env():
    """Test configuration creation from partial environment variables."""
    env_vars = {
        ENV_PROJECT_ID: "env-project",
        ENV_LOCATION_ID: "env-region",
    }

    with patch.dict("os.environ", env_vars, clear=True):
        with pytest.raises(ValidationError) as exc_info:
            BaseConfig.from_env()
        errors = exc_info.value.errors()
        assert any(error["loc"][0] == "key_ring_id" for error in errors)
        assert any(error["loc"][0] == "key_id" for error in errors)

def test_config_initialization():
    """Test direct initialization with valid values."""
    config = BaseConfig(**TEST_CONFIG)
    assert config.project_id == TEST_CONFIG["project_id"]
    assert config.location_id == TEST_CONFIG["location_id"]
    assert config.key_ring_id == TEST_CONFIG["key_ring_id"]
    assert config.key_id == TEST_CONFIG["key_id"]
    assert config.web3_provider_uri == DEFAULT_WEB3_PROVIDER_URI

def test_config_validation_empty():
    """Test validation with empty config."""
    with pytest.raises(ValidationError) as exc_info:
        BaseConfig()
    errors = exc_info.value.errors()
    fields = {"project_id", "location_id", "key_ring_id", "key_id"}
    error_fields = {error["loc"][0] for error in errors}
    assert fields.issubset(error_fields)

def test_config_web3_provider_override():
    """Test setting custom web3 provider URI."""
    custom_uri = "http://custom:8545"
    config = BaseConfig(
        **TEST_CONFIG,
        web3_provider_uri=custom_uri
    )
    assert config.web3_provider_uri == custom_uri

@pytest.mark.parametrize("whitespace", ["  ", "\t", "\n", "\r\n"])
def test_config_whitespace_env_vars(whitespace):
    """Test that whitespace environment variables are treated as missing."""
    env_vars = {
        ENV_PROJECT_ID: whitespace,
        ENV_LOCATION_ID: whitespace,
        ENV_KEY_RING_ID: whitespace,
        ENV_KEY_ID: whitespace
    }

    with patch.dict("os.environ", env_vars, clear=True):
        with pytest.raises(ValidationError) as exc_info:
            BaseConfig.from_env()
        errors = exc_info.value.errors()
        fields = {"project_id", "location_id", "key_ring_id", "key_id"}
        error_fields = {error["loc"][0] for error in errors}
        assert fields.issubset(error_fields)

def test_config_empty_string_validation():
    """Test that empty strings are considered invalid."""
    invalid_config = TEST_CONFIG.copy()
    invalid_config["project_id"] = ""

    with pytest.raises(ValidationError) as exc_info:
        BaseConfig(**invalid_config)
    errors = exc_info.value.errors()
    assert any(error["loc"][0] == "project_id" for error in errors)
